@@ -11,6 +11,9 @@ ONTOUOM = 'http://theworldavatar.com/resource/ontouom/'
 ONTOAGENT = 'http://www.theworldavatar.com/ontology/ontoagent/MSM.owl#'
 ONTODERIVATION = 'https://raw.githubusercontent.com/cambridge-cares/TheWorldAvatar/main/JPS_Ontology/ontology/ontoderivation/OntoDerivation.owl#'
 
+###--- IRI for Geo-Spatial Data ---###
+GEOSPATIAL_LAT_LON_TIME = 'http://www.bigdata.com/rdf/geospatial/literals/v1#lat-lon-time'
+
 ###--- Common PREFIX for SPARQL query ---###
 PREFIX_RDFS = 'PREFIX rdfs:   <' + RDFS_BASE_URL + '> '
 PREFIX_RDF = 'PREFIX rdf:    <' + RDF_BASE_URL + '> '
